from datetime import datetime
from typing import Any
from zoneinfo import ZoneInfo

from django import forms
from django.apps import apps as django_apps
from django.conf import settings
from django.core.exceptions import ObjectDoesNotExist
from edc_form_validators import FormValidator
from edc_screening.utils import get_subject_screening_model_name
from edc_utils import AgeValueError, age
from edc_utils.date import to_utc
from edc_utils.text import convert_php_dateformat


class SubjectConsentFormValidatorMixin(FormValidator):

    """Form Validator mixin for the consent model."""

    subject_screening_model = get_subject_screening_model_name()

    def __init__(self: Any, **kwargs):
        super().__init__(**kwargs)
        self._subject_screening = None
        self._consent_datetime = None
        self.dob = self.cleaned_data.get("dob")
        self.gender = self.cleaned_data.get("gender")
        self.guardian_name = self.cleaned_data.get("guardian_name")
        self.screening_identifier = self.cleaned_data.get("screening_identifier")
        self.tz = ZoneInfo(settings.TIME_ZONE)

    def _clean(self) -> None:
        self.validate_consent_datetime()
        self.validate_age()
        self.validate_gender()
        self.validate_identity()
        super()._clean()

    @property
    def subject_screening_model_cls(self: Any) -> Any:
        return django_apps.get_model(self.subject_screening_model)

    @property
    def consent_datetime(self: Any) -> datetime:
        if not self._consent_datetime:
            if "consent_datetime" in self.cleaned_data:
                if self.add_form and not self.cleaned_data.get("consent_datetime"):
                    raise forms.ValidationError(
                        {"consent_datetime": "This field is required."}
                    )
                self._consent_datetime = to_utc(self.cleaned_data.get("consent_datetime"))
            else:
                self._consent_datetime = self.instance.consent_datetime
        return self._consent_datetime

    @property
    def subject_screening(self: Any) -> Any:
        if not self._subject_screening:
            try:
                self._subject_screening = self.subject_screening_model_cls.objects.get(
                    screening_identifier=self.screening_identifier
                )
            except ObjectDoesNotExist:
                raise forms.ValidationError(
                    'Complete the "Subject Screening" form before proceeding.',
                    code="missing_subject_screening",
                )
        return self._subject_screening

    @property
    def screening_age_in_years(self: Any) -> int:
        """Returns age in years calculated from dob relative to
        screening datetime"""
        try:
            rdelta = age(self.dob, self.subject_screening.report_datetime.date())
        except AgeValueError as e:
            raise forms.ValidationError(str(e))
        return rdelta.years

    def validate_age(self: Any) -> None:
        """Validate age matches that on the screening form."""
        if self.dob and self.screening_age_in_years != self.subject_screening.age_in_years:
            raise forms.ValidationError(
                {
                    "dob": "Age mismatch. The date of birth entered does "
                    f"not match the age at screening. "
                    f"Expected {self.subject_screening.age_in_years}. "
                    f"Got {self.screening_age_in_years}."
                }
            )

    def validate_gender(self: Any) -> None:
        """Validate gender matches that on the screening form."""
        if self.gender != self.subject_screening.gender:
            raise forms.ValidationError(
                {
                    "gender": "Gender mismatch. The gender entered does "
                    f"not match that reported at screening. "
                    f"Expected '{self.subject_screening.get_gender_display()}'. "
                }
            )

    def validate_consent_datetime(self: Any) -> None:
        """Validate consent datetime with the eligibility datetime.

        Eligibility datetime must come first.

        Watchout for timezone, cleaned_data has local TZ.
        """
        if (
            self.consent_datetime - self.subject_screening.eligibility_datetime
        ).total_seconds() < 0:
            local_dt = self.subject_screening.eligibility_datetime.astimezone(self.tz)
            formatted = local_dt.strftime(
                convert_php_dateformat(settings.SHORT_DATETIME_FORMAT)
            )
            raise forms.ValidationError(
                {
                    "consent_datetime": (
                        f"Cannot be before the date and time eligibility "
                        f"was confirmed. Eligibility was confirmed at "
                        f"{formatted}."
                    )
                },
            )

    def validate_identity(self: Any) -> None:
        pass



from .utils import get_consent_model_name, get_reconsent_model_name

consent_codenames = []
for model in [get_reconsent_model_name(), get_consent_model_name()]:
    for action in ["view_", "add_", "change_", "delete_", "view_historical"]:
        consent_codenames.append(f".{action}".join(model.split(".")))


consent_codenames.sort()



from datetime import datetime
from typing import Any

from django import forms
from django.conf import settings
from django.core.exceptions import ObjectDoesNotExist
from edc_visit_tracking.modelform_mixins import get_subject_visit

from ..constants import DEFAULT_CONSENT_GROUP
from ..site_consents import site_consents


class RequiresConsentModelFormMixin:
    def clean(self):
        cleaned_data = super().clean()
        self.validate_against_consent()
        return cleaned_data

    @property
    def report_datetime(self: Any) -> datetime:
        """Returns the report_datetime from directly from
        cleaned_data or via the subject_visit.
        """
        report_datetime = self.cleaned_data.get("report_datetime")
        subject_visit = get_subject_visit(
            self, visit_model_attr=self._meta.model.visit_model_attr()
        )
        if subject_visit and not report_datetime:
            report_datetime = subject_visit.report_datetime
        return report_datetime

    def validate_against_consent(self: Any) -> None:
        """Raise an exception if the report datetime doesn't make
        sense relative to the consent.
        """
        try:
            subject_identifier = self.cleaned_data["appointment"].subject_identifier
        except KeyError:
            subject_visit = get_subject_visit(
                self, visit_model_attr=self._meta.model.visit_model_attr()
            )
            subject_identifier = subject_visit.appointment.subject_identifier
        consent = self.get_consent(subject_identifier, self.report_datetime)
        if self.report_datetime < consent.consent_datetime:
            raise forms.ValidationError("Report datetime cannot be before consent datetime")
        if self.report_datetime.date() < consent.dob:
            raise forms.ValidationError("Report datetime cannot be before DOB")

    @property
    def consent_group(self: Any) -> Any:
        try:
            consent_group = self._meta.model._meta.consent_group
        except AttributeError:
            consent_group = DEFAULT_CONSENT_GROUP
        return consent_group

    @property
    def consent_model(self: Any) -> Any:
        try:
            consent_model = self._meta.model._meta.consent_model
        except AttributeError:
            consent_model = settings.SUBJECT_CONSENT_MODEL
        return consent_model

    def get_consent(self: Any, subject_identifier: str, report_datetime: datetime) -> Any:
        """Return an instance of the consent model"""
        consent_object = site_consents.get_consent(
            report_datetime=report_datetime,
            consent_group=self.consent_group,
            consent_model=self.consent_model,
        )
        try:
            obj = consent_object.model_cls.consent.consent_for_period(
                subject_identifier=subject_identifier, report_datetime=report_datetime
            )
        except ObjectDoesNotExist:
            raise forms.ValidationError(
                f"`{consent_object.model_cls._meta.verbose_name}` does not exist "
                f"to cover this subject on {report_datetime.strftime('Y%-%m-%d %Z')}"
            )
        return obj



from django.contrib import messages
from edc_utils import get_utcnow


def verify_consent(request=None, consent_obj=None):
    consent_obj.is_verified = True
    consent_obj.is_verified_datetime = get_utcnow()
    consent_obj.verified_by = request.user.username
    consent_obj.save(update_fields=["is_verified", "is_verified_datetime", "verified_by"])
    return consent_obj


def unverify_consent(consent_obj=None):
    consent_obj.is_verified = False
    consent_obj.is_verified_datetime = None
    consent_obj.verified_by = None
    consent_obj.save(update_fields=["is_verified", "is_verified_datetime", "verified_by"])
    return consent_obj


def flag_as_verified_against_paper(modeladmin, request, queryset, **kwargs):
    """Flags instance as verified against the paper document."""
    for consent_obj in queryset:
        consent_obj = verify_consent(request, consent_obj)
        messages.add_message(
            request,
            messages.SUCCESS,
            f"'{consent_obj._meta.verbose_name}' for ' "
            f"{consent_obj.subject_identifier}' "
            f"has been verified against the paper document.",
        )


# noinspection PyTypeHints
flag_as_verified_against_paper.short_description = (  # type:ignore
    "Verify consent against paper document"  # type:ignore
)  # type:ignore


def unflag_as_verified_against_paper(modeladmin, request, queryset, **kwargs):
    """Unflags instance as verified."""
    for consent_obj in queryset:
        unverify_consent(consent_obj)


# noinspection PyTypeHints
unflag_as_verified_against_paper.short_description = "Unverify consent"  # type:ignore



from typing import Any

from django import forms
from edc_constants.constants import NO, YES
from edc_screening.utils import get_subject_screening_model_cls


class ConsentModelFormMixinError(Exception):
    pass


class CleanFieldsModelformMixin:
    """A model form mixin calling the default `clean_xxxxx` django
    methods.
    """

    @property
    def subject_screening_model_cls(self):
        return get_subject_screening_model_cls()

    @property
    def subject_screening(self: Any):
        screening_identifier = self.cleaned_data.get(
            "screening_identifier"
        ) or self.initial.get("screening_identifier")
        if not screening_identifier:
            raise ConsentModelFormMixinError(
                "Unable to determine the screening identifier. "
                f"This should be part of the initial form data. Got {self.cleaned_data}"
            )
        return self.subject_screening_model_cls.objects.get(
            screening_identifier=screening_identifier
        )

    def clean_consent_reviewed(self: Any) -> str:
        consent_reviewed = self.cleaned_data.get("consent_reviewed")
        if consent_reviewed != YES:
            raise forms.ValidationError(
                "Complete this part of the informed consent process before continuing.",
                code="invalid",
            )
        return consent_reviewed

    def clean_study_questions(self: Any) -> str:
        study_questions = self.cleaned_data.get("study_questions")
        if study_questions != YES:
            raise forms.ValidationError(
                "Complete this part of the informed consent process before continuing.",
                code="invalid",
            )
        return study_questions

    def clean_assessment_score(self: Any) -> str:
        assessment_score = self.cleaned_data.get("assessment_score")
        if assessment_score != YES:
            raise forms.ValidationError(
                "Complete this part of the informed consent process before continuing.",
                code="invalid",
            )
        return assessment_score

    def clean_consent_copy(self: Any) -> str:
        consent_copy = self.cleaned_data.get("consent_copy")
        if consent_copy == NO:
            raise forms.ValidationError(
                "Complete this part of the informed consent process before continuing.",
                code="invalid",
            )
        return consent_copy

    def clean_consent_signature(self: Any) -> str:
        consent_signature = self.cleaned_data.get("consent_signature")
        if consent_signature != YES:
            raise forms.ValidationError(
                "Complete this part of the informed consent process before continuing.",
                code="invalid",
            )
        return consent_signature

    def clean_initials(self: Any) -> str:
        initials = self.cleaned_data.get("initials")
        if initials and initials != self.subject_screening.initials:
            raise forms.ValidationError(
                "Initials do not match those submitted at screening. "
                f"Expected {self.subject_screening.initials}."
            )
        return initials
